from .Block import Block
from .Directory import Directory
from .btype import FILE, DIR
from .utils import *
def make_children(*child) -> Directory:
    return Directory.make(child)

def unfold(*iter):
    def fn(iter):
        res = Directory(copy=False)
        for cell in iter:
            if hasattr(cell, "__iter__"):
                res += fn(cell)
            else:
                res.append(cell)
        return res
    return fn(iter)




from .btype import __BaseType__

class Abstract:

    def __init__(self, type:__BaseType__, path:str) -> None:
        self.type = type
        self.path = path

    @property
    def __dict__(self):
        return {
            "type": str(self.type),
            "path": self.path
        }


__title__ = 'fileblock'
__description__ = 'simplify file processing'
__version__ = '2.0.2.2022.6.13'
__url__ = "http://github.com/miaobuao/fileblock"
__author__ = 'miaobuao'
__author_email__ = '617802189@qq.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2022 miaobuao'


from .Directory import Directory
from .Block import Block
from .btype import DIR, FILE, __BaseType__

def get_path(children:Directory, force_abs=False):
    """返回Children内所有block的路径"""
    if force_abs:
        res = children.map(lambda x: x.abspath)
    else:
        res = children.map(lambda x: x.path)
    return res

def type_filter(children:Directory, __type:__BaseType__)->Directory:
    def dfs(x):
        return Directory([
            dfs(cell) if hasattr(cell, "__iter__") else cell
            for cell in x
            if hasattr(cell, "__iter__") or (type(cell)==Block and cell.btype==__type)
        ], copy=False)
    return dfs(children)

def file_filter(children:Directory)->Directory:
    """返回Children中的所有FILE类型Block"""
    res = type_filter(children, FILE)
    return res

def dir_filter(children:Directory)->Directory:
    """返回Children中的所有DIR类型Block"""
    res = type_filter(children, DIR)
    return res

def extension_filter(dir:Directory, *extensions):
    def dfs(x: Directory):
        return Directory([
            dfs(cell) if hasattr(cell, "__iter__") else cell
            for cell in x
            if hasattr(cell, "__iter__") or (type(cell)==Block and cell.extension in extensions)
        ], copy=False)
    return dfs(dir)

def remove(target:Directory):
    if type(target) == Directory:
        target.map(lambda x: x.remove())
    elif type(target) == Block:
        target.remove()
    else:
        raise ValueError(target," cannot be removed!")


class __BaseType__:
    
    def __init__(self, name) -> None:
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.__str__()
    
    def __eq__(self, __o) -> bool:
        if type(__o) == str:
            return __o == self.name
        elif type(__o) == __BaseType__:
            return __o.name == self.name
        return False


FILE = __BaseType__("File")
DIR = __BaseType__("Dir")

if __name__ == "__main__":
    print(FILE.__dict__)


from os.path import split, splitext, isfile, exists, join, isdir, abspath
from os import listdir, makedirs, mkdir, remove, rmdir
from .Directory import Directory
from .Abstrcat import Abstract
from .btype import FILE, DIR
import shutil


class Block:

    def __init__(self, path: str):
        '''
            @parameter path - 文件/文件夹路径
        '''
        self.__path = path
    
    def sub_block(self, path: str):
        '''
            @parameter path - 子文件或子文件夹的相对路径
            @returns
                子文件/子文件夹的Block对象
        '''
        sub_path = self.join_path(path)
        return Block(sub_path)
    
    def append(self, name, type=FILE):
        '''
            往block中新增一个文件/文件夹

            TIPS:
                当type == FILE时，若name形如x1/x2, 则会新建x1文件夹，返回的是x2的Block对象，而不是x1
        '''
        path = self.join_path(name)
        if type == FILE:
            dir_name, _ = split(path)
            if not exists(dir_name):
                makedirs(dir_name)
            with open(path, 'w+') as f:
                pass
        elif type == DIR:
            makedirs(path)
        return Block(path)

    def join_path(self, path: str)->str:
        '''
            @parameter path - 需要拼接的文件/文件夹路径
            @returns:
                拼接后的路径
        '''
        return join(self.path, path)
    
    def remove(self):
        '''
             如果该block对象是真实存在的文件/文件夹, 则从磁盘上永久删除该文件/文件夹
        '''
        if self.exists:
            if self.isfile:
                remove(self.path)
            else:
                rmdir(self.path)
        else:
            raise FileNotFoundError("文件/文件夹不存在时，无法执行block.remove()")

    def moveTo(self, target):
        if type(target) == str:
            target = Block(target)
        if not self.exists:
            raise FileNotFoundError(self.path, " not found! ")
        if not target.exists:
            makedirs(target.path)
        if not target.isdir:
            raise Exception(target.path, " is not a dir.")
        shutil.move(self.path, target.path)
        return self
        
    def copyTo(self, target):
        if type(target) == str:
            target = Block(target)
        if not self.exists:
            raise FileNotFoundError(self.path, " not found! ")
        if not target.exists:
            makedirs(target.path)
        if not target.isdir:
            raise Exception(target.path, " is not a dir.")
        mask_len = len(self.abspath)
        if self.isfile:
            path = target.join_path(self.file_full_name)
            with open(path, 'wb+') as f:
                f.write(self.get_file_contents())
        else:
            for leaf in self.leaves:
                path = target.join_path("." + leaf.abspath[mask_len:])
                dir_path, name = split(path)
                if leaf.isfile:
                    if not exists(dir_path):
                        makedirs(dir_path)
                    with open(path, 'wb+') as f:
                        f.write(leaf.get_file_contents())
                else:
                    if not exists(path):
                        makedirs(path)
        return self

    def search(self, name, type=None):
        '''
            在block内进行搜索，返回所有匹配的文件/文件夹
            
            可选参数type的值应为 fileblock.FILE 或 fileblock.DIR
        '''
        leaves = self.leaves
        return Directory([
            leaf for leaf in leaves
            if name in split(leaf.path)[1] and ((type is not None and leaf.btype==type) or type is None)
        ], copy=False)
    

    def cut(self, *rates):
        '''
            切割文件夹内的文件块
            @parameter - rates 分割比例，可以是整数，也可以是浮点数
        '''
        if not rates:
            rates = (1, )
        if not self.isdir:
            return None
        children = self.children
        gross = sum(rates)
        n = children.__len__()
        cell = n / gross
        acc = 0
        res = Directory(copy=False)
        rates_len = len(rates)
        for i in range(len(rates)):
            if rates_len == i + 1:
                res.append(children[acc : ])
            else:
                idx = int(rates[i] * cell)
                res.append(children[acc : acc + idx])
                acc += idx
        return res
    
    def get_file_contents(self, mode='rb', encoding=None):
        if self.isfile:
            with open(self.path, mode, encoding=encoding) as f:
                d = f.read()
            return d
        raise TypeError("This block is not a file-type block")

    def enter(self, path):
        self.__path = self.join_path(path)
        return self

    @property
    def children(self):
        res = Directory()
        if self.isdir:
            dirs = listdir(self.path)
            for dir in dirs:
                res.append(self.sub_block(dir))
        return res

    @property
    def path(self):
        return self.__path
    
    @property
    def leaves(self):
        '''
        返回文件树中的所有叶子节点
        '''
        children = self.children
        if len(children) > 0:
            res = Directory()
            for child in children:
                res.extend(child.leaves)
            return res
        else:
            return Directory([self])

    @property
    def abspath(self):
        return abspath(self.path)
    
    @property
    def directory(self):
        return split(self.path)[0]
    
    @property
    def file_full_name(self):
        return split(self.path)[1]
    
    @property
    def filename(self):
        return splitext(self.file_full_name)[0]
    
    @property
    def extension(self):
        return splitext(self.file_full_name)[1]
    
    @property
    def abstract(self, force_abspath=False):
        return Abstract(self.btype, self.abspath if force_abspath else self.path)

    @property
    def isfile(self) -> bool:
        '''
            block类型是文件则返回True
        '''
        return isfile(self.path)

    @property
    def isdir(self) -> bool:
        '''
            block类型是文件夹则返回True
        '''
        return isdir(self.path)
    
    @property
    def exists(self) -> bool:
        '''
            该block是否存在
        '''
        return exists(self.path)
    
    @property
    def super_dir_name(self):
        tmp = split(self.path)[0]
        return split(tmp)[1]

    @property
    def btype(self):
        return FILE if self.isfile else DIR if self.isdir else None
    
    def __str__(self):
        # return "(%s block: %s)" % (self.btype, self.path)
        return self.path
    
    def __repr__(self) -> str:
        return self.__str__()


if __name__ == "__main__":

    b = Block("../")
    print(b.children)


from random import random
import json
from . import Block
from .btype import __BaseType__

def deep_maker(x):
    if hasattr(x, "__iter__"):
        res = []
        for cell in x:
            res.append(deep_maker(cell))
        return Directory(res, copy=False, deep_make=False)
    else:
        return x


class Directory:

    def __init__(self, data=[], copy=True, deep_make=False):
        if isinstance(data, Directory):
            data = data.data
        if deep_make:
            data = deep_maker(data)
        if copy:
            self.data = data.copy()
        else:
            self.data = data


    def map(self, fn):
        def dfs(x):
            return Directory([
                dfs(cell) if isinstance(cell, Directory) else fn(cell)
                for cell in x
            ], copy=False, deep_make=False)
        return dfs(self)


    def to_json(self, fpath: str, file_only=False, dir_only=False, abspath = False, indent=None, encoding="utf8"):
        """
            注: 若file_only 和 dir_only 同时为 True 则 全都输出.

            tips: if `file_only` == `dir_only` == True,
                    then output is all.
        """
        def convert(child):
            if type(child) == Directory:
                res = []
                for c in child:
                    tmp = convert(c)
                    if tmp:
                        res.append(tmp)
                return res
            if file_only and not dir_only:
                if child.isfile:
                    return child.abstract(abspath).__dict__
            elif dir_only and not file_only:
                if child.isdir:
                    return child.abstract(abspath).__dict__
            else:
                return child.abstract(abspath).__dict__

        data = convert(self)
        with open(fpath, "w+", encoding=encoding) as f:
            json.dump(data, f, indent=indent, ensure_ascii=False)
        

    def unfold(self):
        def proc(children, out):
            if isinstance(children, Directory):
                for cell in children:
                    proc(cell, out)
            else:
                out.append(children)
        res = []
        proc(self.data, res)
        return Directory(res, copy=False, deep_make=False)


    def copy(self):
        return Directory(self.data, copy=True, deep_make=False)


    def shuffle(self):
        res = self.copy()
        le = res.__len__()
        for i in range(1, le+1):
            idx = int(random() * (le - i))
            res[idx], res[le - i] = res[le - i], res[idx]
        return res
    

    def append(self, obj):
        self.data.append(obj)
        return self


    def extend(self, obj):
        self.data.extend(obj)
        return self
    

    def remove(self, x):
        self.data.remove(x)
        return self


    def pop(self, idx):
        return self.data.pop(idx)
    
    def extension_filter(self, *extensions):
        rep_ext = [ 
            '.' + e
            for e in extensions
            if e[0] != '.'
        ]
        def dfs(x: Directory):
            return Directory([
                dfs(cell) if hasattr(cell, "__iter__") else cell
                for cell in x
                if hasattr(cell, "__iter__") or
                    (type(cell) == Block.Block and
                    (cell.extension in extensions or cell.extension in rep_ext))
            ], copy=False, deep_make=False)
        return dfs(self)
            

    @staticmethod
    def make(*child, copy=True, deep_make=True):
        def proc(children):
            if hasattr(children[0], "__iter__"):
                res = Directory(copy=False, deep_make=False)
                for child in children:
                    res.append(proc(child))
                return res
            return Directory(children, copy=copy, deep_make=deep_make)
        return proc(child)

    @property
    def abspaths(self):
        return Directory([child.abspath for child in self], copy=False, deep_make=False)
    
    @property
    def super_dir_names(self):
        return self.map(lambda x: x.super_dir_name)

    
    def __add__(self, x):
        return Directory(self.data + x, copy=False, deep_make=False)
    

    def __len__(self):
        return self.data.__len__()
    

    def __iter__(self):
        return self.data.__iter__()
    
    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return Directory(self.data[idx], copy=False, deep_make=False)
        return self.data[idx]

    def __setitem__(self, k, v):
        self.data[k] = v
        return self

    def __str__(self) -> str:
        return str(self.data)
    

    def __repr__(self) -> str:
        return self.__str__()

if __name__  == "__main__":

    c = Directory([1, 2, 3])
    x = c + Directory([2, 3, 4])
    print(x)
    